package MonitoringFramework.ACLClass;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicLong;

import MonitoringFramework.classtrn.BlackBoard;
import MonitoringFramework.classtrn.Constants;
import MonitoringFramework.classtrn.LogManager;
import MonitoringFramework.classtrn.PrintToScreen;
import MonitoringFramework.classtrn.ScenarioParameter;
import MonitoringFramework.classtrn.WrongTriggerException;
import Scenario.Utility.StringUtils;

public abstract class ACLScenario implements Runnable, ACLEventListener, ACLEventProcessor{
	
	protected final long scenarioID = NEXT_ID.getAndIncrement();
	protected static AtomicLong NEXT_ID = new AtomicLong(0);
	
	
	protected String terminateEvent="";
	protected ACLContract ownerContract;
	protected ACLEventType triggeringEvent;//to be set when the scenario is instantiated//
	protected Queue<ACLEventType> recievedEvents;
	protected Queue<ACLEventType> acceptedEvents; //let's see if I can do without this queue
	protected EventProcessorState scenarioState;
	//define a list of task for each scenario instance
	protected List<ACLEventProcessor> scenarioTaskList;
	protected Map<String, List<ScenarioParameter>> parameterMap; //Let's just keep it concurrent for now. The notifier thread accesses it and then the scenario thread will read
	//but it is unlikely that it will be simultaneous accessed. 
	
	
	

	protected ACLScenario(ACLContract ownerContract, ACLEventType trigger) {
		if(!trigger.getEventName().equalsIgnoreCase(getExpectedTriggeringEventName())){
			PrintToScreen.printToScreen("trigger expected"+ getExpectedTriggeringEventName()+" but recieved: "+ trigger.getEventName());
			LogManager.getInstance().error(ownerContract,"trigger expected"+ getExpectedTriggeringEventName()+" but recieved: "+ trigger.getEventName());			
		}
		this.ownerContract = ownerContract;
		this.triggeringEvent  = trigger;
		this.recievedEvents = new ConcurrentLinkedQueue<>();	//each scenario instance uniquely receives the event.i.e. scenario instances will not share the received event queue
		this.acceptedEvents = new ConcurrentLinkedQueue<>(); 
		this.scenarioState= EventProcessorState.CREATED;//not really needed to be concurrent.
		if(scenarioHasParameterizedEvents()){
			parameterMap = new ConcurrentHashMap<>(); //Put only the events which have parameters associated with them!!
			initializeParameterMap(); //for now I am tying all the event parameters to the triggering event parameter
			try {
				populateParameterMapFromEvent(trigger);
			} catch (WrongTriggerException e) {
				PrintToScreen.printToScreen(e.getMessage());
				e.printStackTrace();
			}
		}
		initializeTasksForScenario(); //task creation happens during this call.
		//set the first task to set state 
		if(!getConstituentsEventProcessors().isEmpty()){
			getConstituentsEventProcessors().get(0).setUpToAcceptEvents();
		}
				
	}
	
	/**
	 * Abstract Methods.
	 */
	protected abstract void initializeTasksForScenario();
	protected abstract Boolean scenarioHasParameterizedEvents();
	protected abstract Map<String, List<ScenarioParameter>> getParameterMap();
	//protected abstract List<String> getExpectedEventNames();
	protected abstract String getExpectedTriggeringEventName();
	protected abstract void initializeParameterMap();
	
	
	public String getTerminateEvent() {
		return terminateEvent;
	}
	
	/**
	 *  Must be called when a new scenario instance is created
	 * @param terminateEvent
	 */
	public void setTerminateEvent(String terminateEvent) {
		this.terminateEvent = terminateEvent;
	}

	public ACLContract getOwnerContract() {   
		return ownerContract;
	}

	private void setOwnerContract(ACLContract ownerContract) { //don't expose this unnecessarily.
		this.ownerContract = ownerContract;
	}

	public ACLEventType getTriggeringEvent() {
		return triggeringEvent;
	}
	
	/**
	 * Must be called when a new scenario instance is created
	 * @param triggeringEvent
	 */
	public void setTriggeringEvent(ACLEvent triggeringEvent) {
		this.triggeringEvent = triggeringEvent;
	}
	
	@Override
	public List<ACLEventProcessor> getConstituentsEventProcessors() {
		if(scenarioTaskList==null){
			scenarioTaskList = new ArrayList<>(); // Task list is always a simple list.
			//A dynamic task may have other dynamic task in it's list and that may be a skip list 
		}
		return scenarioTaskList;
	}
	
	@Override
	public ACLEventListenertype getACLEventListenerType() {
		return ACLEventListenertype.ACLSCENARIO;
	}
	
	
	public void notifyOfACLEvent(final ACLEventType event) {
		//Upon notification just make a copy of the event and put it on your Queue and immediately return 
		//Process the event from inside of the run method
		//Not making a copy for now. Once created events are never modified.
		recievedEvents.add(event);
	}
	
	@Override
	public void registerForACLEvents() {
		for(String event: getExpectedEvents()){
			BlackBoard.registerForACLEventNotification(event, this);
		}		
	}

	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((ownerContract == null) ? 0 : ownerContract.hashCode());
		result = prime * result + (int) (scenarioID ^ (scenarioID >>> 32));
		return result;
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof ACLScenario)) {
			return false;
		}
		ACLScenario other = (ACLScenario) obj;
		if (ownerContract == null) {
			if (other.ownerContract != null) {
				return false;
			}
		} else if (!ownerContract.equals(other.ownerContract)) {
			return false;
		}
		if (scenarioID != other.scenarioID) {
			return false;
		}
		return true;
	}


	public long getId(){
		return this.scenarioID;
	}
	
	//let's just allow the new to be called.
	//public abstract ACLScenario createScenarioInstance(ACLContract ownerContract,ACLEvent trigger) throws WrongTriggerException;
	
	@Override
	public EventProcessorState getEventProcesorState(){
		if(!this.scenarioState.equals(EventProcessorState.CREATED)){
			return this.scenarioState;
		}
		else{			
			//A scenario's state is cumulative of it's task's state. It is turned to PASS if all the tasks are in PASSED state 
			if(this.getConstituentsEventProcessors().stream().allMatch(task-> task.getEventProcesorState().equals(EventProcessorState.PASSED))){
				this.scenarioState = EventProcessorState.PASSED;
				System.out.println("Scenario " + this.getListenerName()+" InstanceId: "+ this.getListenerId() + " is successful!");
			}
			//if any of the scenario task is in 'fail' state then the Scenario is considered to have failed.
			if(!this.scenarioState.equals(EventProcessorState.PASSED)){ //once passed do not set it fail. just being paranoid here.
				if(this.getConstituentsEventProcessors().stream().anyMatch(task -> task.getEventProcesorState().equals(EventProcessorState.FAILED))){
					System.out.println("Setting the scenario state to fail because some of the constituents are in failed state");
					this.scenarioState = EventProcessorState.FAILED;
				}
			}
		}
		return this.scenarioState;
		
	}
	
	protected Boolean checkParameterAndAcceptEvent(ACLEventType event){
		if(scenarioHasParameterizedEvents()){
			return doParametersMatch(event);
		}
		else{
			return true;//accept all the events if scenario doesn't have parameters.
		}
	}
	
	private Boolean doParametersMatch(ACLEventType event) {
		String eventName= event.getEventName();
		//Get the parametermap and look for the event if found compare the parameter value. 
		//If the event is not in the parameterMap then return true because the expected event is parameter-less
		if(getParameterMap().containsKey(eventName)){
			List<ScenarioParameter> scenarioParametersForTheEventFromMap = getParameterMap().get(eventName);			 
			List<Object> expectedParameterValues = new ArrayList<>();
			scenarioParametersForTheEventFromMap.forEach((scenarioParameter) -> expectedParameterValues.add(scenarioParameter.getParameterValue())); //Not using the streams. order has to be preserved.
			//now retrieve the actual value from the event.
			if(event.hasParameter()){
				List<Object> actualParameterValues = ((ParameterizedACLEvent)event).getParameterValue();
				if(actualParameterValues==null || actualParameterValues.isEmpty()){
					//something went wrong log it.
					PrintToScreen.printToScreen("Scenario: "+ getListenerName()+ " instanceid: "+ getListenerId()+" could not resolve the parameter value for event: "+ event.getEventName() );
				}				
				//PrintToScreen.printToScreen("Expected ParameterValue: "+ expectedParameterValue+" actual ParameterValue: " + actualParameterValue+ " for event: "+ eventName);
				Boolean result = areParameterValuesTruelyEqual(actualParameterValues, expectedParameterValues);
				if(!result){
					//log if rejected.
					PrintToScreen.printErrorToScreen("Expected ParameterValues: "+ expectedParameterValues.toString()+" actual ParameterValue: " + actualParameterValues.toString()+ " for event: "+ eventName);
				}
				return result;//actualParameterValues.equals(expectedParameterValues);
			}
			else{
				//if we are here that means that a non-parameterized event was fired when a parameterized event was expected.
				return false;
			}
		}
		else{
			return true;
		}		
	}
	
	private Boolean areParameterValuesTruelyEqual(List<Object>actualParameters, List<Object> expectedParameters){
		// actual parameters are in the list because event may be fired like so - Fire(E1(value1, value2)). 
		//The type of Values may be be - Integer, String, ContractInstance.//Whereas the elements in the list of expectedParameters may itself be a list 
		//because it may be associated with an observability which returns a list. 
		//At this level we just want to check that such a list contains the parameter.
		try{
			for (int i = 0; i < expectedParameters.size(); i++) {
				Object expectedParameter = expectedParameters.get(i);
				Object actualParameter = actualParameters.get(i);
				//if the expected parameter is a list
				if(expectedParameter instanceof List<?>){
					//check if the list contains the actual parameter
					if(!((List<?>)expectedParameter).contains(actualParameter)){
						return false;
					}
				}
				else{
					if(!expectedParameter.equals(actualParameter)){
						return false;
					}
				}
			}	
		}
		catch(Exception e){
			PrintToScreen.printErrorToScreen("Exception occured while checking parameter equivalency: "+ e.getMessage());
			e.printStackTrace();
			return false;
		}			
		return true;		
	}
	
	/**
	 * Takes the parameters from the event and sets the values in the parameter-map for the event name.
	 * @param event
	 */
	protected void populateParameterMapFromEvent(ACLEventType event) throws WrongTriggerException{
		List<Object> triggerEventParameterValues=null;
		String eventName = event.getEventName();
		if(event.hasParameter()){
			//open up the event.//set up the map 
			triggerEventParameterValues =  ((ParameterizedACLEvent)event).getParameterValue();			
			List<ScenarioParameter> scenarioParametersForTheEventFromMap = getParameterMap().get(eventName);
			if(triggerEventParameterValues.size() != scenarioParametersForTheEventFromMap.size()){
				throw new WrongTriggerException(this.getListenerName(), "Mismatch in the expected and actual number of parameters for the triggering event");
			}
			for (int i = 0; i <scenarioParametersForTheEventFromMap.size() ; i++) {
				Object parametervalueFromFiredEvent = triggerEventParameterValues.get(i); //not really checking for number of parameters to match!
				scenarioParametersForTheEventFromMap.get(i).setValue(parametervalueFromFiredEvent);
			}
			//getParameterMap().get(eventName).setValue(triggerEventParameterValue); //This is how to do it if there was only one parameter with all the events. 
		}
	}
	
	@Override
	public void setProcessorStatusToFail(){
		System.out.println("Setting the scenario state to fail: in ACL scenario class");
		this.scenarioState = EventProcessorState.FAILED;
	}	
	

	@Override
	public long getListenerId() {
		return scenarioID;
	}
	
	
	@Override
	public void run() {
		/**
		 * Steps - 
		 * 1. Register with Blackboard for all the events that are pertinent to this scenario.
		 * 2.The executor service starts up- picks up events one by one from the firedEventQueue.
		 * 3. The events that put in firedEvent queue have to be verified - 
		 * a. If it is parameter-less even put it in the fired event  queue right away
		 * b. If it is parameterized event- parameter should be matched   
		 */
		/**
		 * Call start on this thread when the scenario state is Created.
		 * Check to see if any event has been recieved.The 
		 */
		
		do {
			if(!recievedEvents.isEmpty()){
				//dequeue the event and check to see if it can be accepted and put on the accepted queue.
				ACLEventType recievedevent = recievedEvents.poll();
				if(checkParameterAndAcceptEvent(recievedevent)){
					acceptedEvents.add(recievedevent);
				}
			}
			//flood the task list with the accepted event but before doing that Now is our chance to check 
			while(!acceptedEvents.isEmpty()){ //accepted events being modified from only this thread so it is safe.				
				Boolean didCurrentTaskSucceed = false;
				for(int i=0; i<getConstituentsEventProcessors().size() ; i++){
					int nextTaskIndex = i+1;
					ACLEventProcessor currentTask = getConstituentsEventProcessors().get(i);
					didCurrentTaskSucceed = currentTask.getEventProcesorState().equals(EventProcessorState.PASSED);
					if(didCurrentTaskSucceed && nextTaskIndex < getConstituentsEventProcessors().size() ){ //The first task at 0 indices must have been set up to accept events already
						ACLEventProcessor nextTask = getConstituentsEventProcessors().get(nextTaskIndex);
						if(nextTask.getEventProcesorState().equals(EventProcessorState.CREATED)){
							nextTask.setUpToAcceptEvents();
						}
					}					
				}
				//Now everyone receives the event. If task hasn't been set up and is receiving the event (that it should receive) 
				ACLEventType acceptedEvent = acceptedEvents.poll();
				getEventProcesorState(); //calling getScenarioState so that any tasks that can be set to pass, set themselves to pass
				if(acceptedEvent!=null && StringUtils.isNotEmpty(acceptedEvent.getEventName())){ //unnecessary check but I will keep it just to be sure. 
					this.getConstituentsEventProcessors().forEach((task) -> task.processEvent(acceptedEvent));
				}
				
				//sSystem.out.println(acceptedEvent);
			}
			//check if it's the end of the eventNotification ands
			//System.out.println(BlackBoard.getInstance().isTimeToshutDownEventNotification());
			if(isItTheEnd()){
				getEventProcesorState();
				if(this.scenarioState.equals(EventProcessorState.CREATED)){
					PrintToScreen.printToScreen("Scenario being set to failed because some expected events did not occur");
					this.setProcessorStatusToFail();
				}					
			}
			try {
				Thread.sleep(1000 * Constants.SCENARIOTHREADSLEEPTIMEINSECOND); //This is random 20 second sleep. Need to have better strategy for this.
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			
		} while (this.scenarioState.equals(EventProcessorState.CREATED));//run until scenario either has passed or failed.
		
	}
	
	
	private boolean isItTheEnd(){
		//@zhibo
		return BlackBoard.getInstance().isTimeToshutDownEventNotification() && recievedEvents.isEmpty() && acceptedEvents.isEmpty();
		//return recievedEvents.isEmpty() && acceptedEvents.isEmpty();
	}

	/* (non-Javadoc)
	 * @see classtrn.ACLEventProcessor#setUpToAcceptEvents()
	 */
	@Override
	public void setUpToAcceptEvents() {
		//Scenario instances do not have to be set up to receive events. They can receive events as soon as they are created.
		//Just a place holder.
	}
	
	@Override
	public void processEvent(ACLEventType event){
		//place holder.
		//The running thread (run method) process the events instead.

		
	}

	/* (non-Javadoc)
	 * @see classtrn.ACLEventProcessor#getParentEventProcessor()
	 */
	@Override
	public ACLEventProcessor getParentEventProcessor() {
		/*this shouldn'e be called. A scenario doesn't have a parent event processor. 
		//It does have an owner contract but contracts are not event processors (they just receive events, so they are eventListneres)*/
		return null;
	}

	/* (non-Javadoc)
	 * @see classtrn.ACLEventProcessor#skip()
	 */
	@Override
	public void skip() {
	//Place holder. Scenario can't be skipped...For now.
		
	}

	/* (non-Javadoc)
	 * @see classtrn.ACLEventProcessor#getForbiddenEvents()
	 */
	@Override
	public Set<String> getForbiddenEvents() {
		/*this shouldn'e be called. A scenario doesn't have forbidden events. forbidden events are defined at the eventExpression level, which get accumalated to tasks.
		 * So tasks and eventExpressions have forbidden events*/
		return null;
	}

}
