package MonitoringFramework.classtrn;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import MonitoringFramework.ACLClass.ACLEventListener;
import MonitoringFramework.ACLClass.ACLEventType;

public class BlackBoard {
	
	 private static BlackBoard _instance = null;
	 private Integer contractCount = 50; //need some intelligent mechanism for this.
	 private Integer ScenarioCount = 150;
	
	// a list of all the possible ACL event names
	// To be set once during startup. Not Threadsafe for writes.
	// Reads are good.
	private List<String> allPossibleEvents;
	// Queue of fired events
	private Queue<ACLEventType> firedEvents;	
	
	private Map<String, List<ACLEventListener>> eventListeners;
	
	private AtomicBoolean iutShutdownNotification; //should come from MainContract?
	
	private Object eventqueueLock= new Object();
	
	public List<String> getAllPossibleEvents() {
		return allPossibleEvents;
	}
	public void setAllPossibleEvents(List<String> allPossibleEvents) {
		BlackBoard.getInstance().allPossibleEvents.addAll(allPossibleEvents);
	}
	
	private Queue<ACLEventType> getFiredEvents() {
		return firedEvents;
	}
	
	/**
	 * Do not call this from constructor. if you do, you are a bad person.
	 */
	public static void registerForACLEventNotification(String eventName, ACLEventListener eventListenter){
		List<ACLEventListener> existingEventListeners = BlackBoard.getInstance().getEventListeners().get(eventName);
		if(existingEventListeners!=null){
			//is it exists that means a synchronized list was put in, in earlier attempts. Now modify the list
			//If other calling thread has got a lock then code waits for that execution to complete otherwise continues.
			//If the list is being modified then reads (and iterations synchronized separately) will have to wait for this operation complete as wells
			existingEventListeners.add(eventListenter);
		}
		else{
			//use carefully when iterating. While iterating it would need to lock the whole list. i.e. would need to synchronize on list object itself
			List<ACLEventListener> newEventListnsers =Collections.synchronizedList(new ArrayList<>()); 
			newEventListnsers.add(eventListenter);			
			BlackBoard.getInstance().getEventListeners().put(eventName,newEventListnsers );
		}
	}
	
	/**
	 * This MUST be called when finalize is called upon contract instances or when scenarios terminate so that the scenario instances become eligible for 
	   garbage collection
	   Potential deadlock nominee - I will not call the de-register rightaway- on being notified about the terminating event, scenario must return immediately and make this call from 
	   other method i.e. not from the inside of notify. 
	 * @param eventName
	 * @param eventListenter
	 */
	public static void deRegister(String eventName, ACLEventListener eventListenter){
		List<ACLEventListener> existingEventListeners = BlackBoard.getInstance().getEventListeners().get(eventName);
		if(existingEventListeners!=null){
			//is it exists that means a synchronized list was put in, in earlier attempts. Now modify the list
			//If other calling thread has got a lock then code waits for that execution to complete otherwise continues.
			//If the list is being modified then reads (and iterations synchronized separately) will have to wait for this operation complete as wells
			existingEventListeners.remove(eventListenter);
		}
		else{
			LogManager.getInstance().error(eventListenter, "Listener tried to de-register without registering. Possible race conditions!");
		}
	}
	
	public static void deRegisterFromAllTheEvents(ACLEventListener eventListenter){
		
	}
	
	
	public Map<String, List<ACLEventListener>> getEventListeners() {
		return eventListeners;
	}
	
	public static BlackBoard getInstance () {     
	      if (_instance == null) {
	         synchronized (BlackBoard.class) {
	            if (_instance == null) {               
	               _instance = new BlackBoard();
	            }}}
	      return _instance;
	   }
	 
	 protected BlackBoard () {
		 firedEvents = new ConcurrentLinkedQueue<>();	
	      allPossibleEvents = new ArrayList<>();
	      iutShutdownNotification = new AtomicBoolean(false);//initialize with false.
	      eventListeners = new ConcurrentHashMap<String, List<ACLEventListener>>((contractCount+ScenarioCount)*50); //Setting initial capacity. Should ideally be bound to test data size
	      Thread eventNotifierThread = new Thread(new ACLEventNotifier());
	      eventNotifierThread.setName("BlackBoardEventNotifier");
	      eventNotifierThread.start();
	   }
	
	private class ACLEventNotifier extends Observable implements Runnable {

		@Override
		public void run() {
			/*Takes the event out of the firedEvents queue
			 *  If any scenario needs to be instantiated - they need to be instantiated and registered for event listening 
			 *  Then all the registered scenario for this event should be notified 
			 * */
			 //poll is atomic. shouldn't cause threading issues. 
			while(!isTimeToshutDownEventNotification()){
				//System.out.println("looking for events");
				ACLEventType currentEvent = getFiredEvents().poll();
				if(currentEvent!=null){
					//Found an event. Do stuff
					String eventName = currentEvent.getEventName();
					System.out.println("Found Event: "+ currentEvent.getEventName());
					//Look for the registered listener for this event name and dispatch.
					List<ACLEventListener> existingEventListeners = BlackBoard.getInstance().getEventListeners().get(eventName);
					if(existingEventListeners!=null){
						//get a lock for iteration
						//if run into concurrentmodificatiion exception gonna try with iterator first and then try the concurrent implementation.
						synchronized (existingEventListeners) {
							for (ACLEventListener aclEventListener : existingEventListeners) {
								//lets just put the call to get scenario status here. So before firing the event we ask the eventListener 
//								if(aclEventListener.getACLEventListenerType().equals(ACLEventListenertype.ACLSCENARIO)){
//									System.out.println(((ACLScenario) aclEventListener).getScenarioState());
//								}
								aclEventListener.notifyOfACLEvent(currentEvent); /*The contract or scenario can not register(or add a new listener) for the event that they are 
								being notified about*/
								System.out.println("Notified listener: with name: " + aclEventListener.getListenerName()+ " and with id: "+ aclEventListener.getListenerId()+" of type: "+ aclEventListener.getACLEventListenerType()
								+" of  Event: "+ currentEvent.getEventName());
							}
						}						
					}
					else{
						if(!BlackBoard.getInstance().iutShutdownNotification.get()){
							// This is dubious. I am putting the event back on the queue if there were no listeners. So it can be observed later.
							//Only put it back if iut is still running.
							//@zhibo how to handle situationis interesting, because some times dead loops are happenning if no one listening to this event, while this event is the only one in the queue
							PrintToScreen.printToScreen("No listener found..putting the event"+currentEvent.getEventName()+ "back on queue");
							getFiredEvents().add(currentEvent);
						}
					}
				}
				else{
					try {
						Thread.sleep(1000 * Constants.BLACKBOARDTHREADSLEEPTIMEINSECOND); //This is random 10 second sleep. Need to have better strategy for this.
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}//sleep for 5 seconds. Its random 
				}
			}
			//if it comes here that means event notification has been shut down. No more events coming through.

		}

	}
	
	/**
	 * No more responsibility should execute after this.
	 */
	public void notifyIUTExit(){		
		this.iutShutdownNotification.compareAndSet(false, true);
	}
	
//	public boolean hasIUTBeenShutDown(){
//		return this.iutShutdownNotification.get();
//	}
	
	public Boolean isTimeToshutDownEventNotification(){
		if(this.iutShutdownNotification.get()){
			//check if the queue is empty but first get a lock on queue synchronizing object because isEmpty() is not an atmoic operation.
			synchronized (eventqueueLock) {
				return getFiredEvents().isEmpty();
			}
		}
		else{
			//IUT shut down notification has not come, keep processing.
			return false;
		}
		//return this.iutShutdownNotification && getFiredEvents().isEmpty(); //don;t shut down the event notification until you have processed all the fired events.
	}  
	
	
	public void fireAnEvent(ACLEventType event){
		if(this.isTimeToshutDownEventNotification()){
			synchronized (eventqueueLock) { // This lock is always available before the IUT shutdown.
				this.getFiredEvents().add(event);
			}	
		}
		else{
			//fire away. No need to get a lock.
			this.getFiredEvents().add(event);
		}
			
	}
	

}
